# Licensed under the GNU Lesser General Public License v3.0.
# ezudesign Copyright (C) 2023 numlinka.

__all__ = [
    "configuration",
    "eventhub",
    "taskflow",
    "taskpool",
    "tasksequence"
]

__name__ = "ezudesign"
__author__ = "numlinka"
__license__ = "LGPLv3"
__copyright__ = "Copyright (C) 2023 numlinka"

__version_info__ = (0, 2, 0)
__version__ = ".".join(map(str, __version_info__))

# internal
from . import configuration
from . import eventhub
from . import taskflow
from . import taskpool
from . import tasksequence



# Licensed under the GNU Lesser General Public License v3.0.
# ezudesign Copyright (C) 2023 numlinka.

# std
import time
import threading

from typing import Callable, Iterable, Mapping, Optional, Any, NoReturn
from dataclasses import dataclass

# site
from typex import MultitonAtomic


TASK_SEQUENCE_START = "task_sequence_start"
TASK_SEQUENCE_ADD = "task_sequence_add"
TASK_SEQUENCE_EXEC = "task_sequence_exec"
TASK_SEQUENCE_IDLE = "task_sequence_idle"
TASK_SEQUENCE_EXCEPT = "task_sequence_except"
TASK_SEQUENCE_CLEAR = "task_sequence_clear"


@dataclass
class SyncTaskUnit (object):
    iid: int
    name: str
    task: Callable
    args: Iterable[Any]
    kwargs: Mapping[str, Any]


class TaskSequence (object):
    def __init__(self):
        self._lock = threading.RLock()
        self.thread: Optional[threading.Thread] = None
        self._event = threading.Event()
        self.__atomic = MultitonAtomic(instance_name="numlinka - ezudesign - TaskSequence")
        self.__task_list: list[SyncTaskUnit]

    @property
    def running(self) -> bool:
        return self.thread is not None and self.thread.is_alive()

    def callback(self, action: str, detail: Optional[SyncTaskUnit | Exception] = None) -> None:
        ...

    def add_task(self, name: str, task: Callable, args: Iterable[Any], kwargs: Mapping[str, Any]) -> int:
        if not isinstance(name, str):
            raise TypeError(f"Expected `name` to be str, but got {type(name)}.")

        if not callable(task):
            raise TypeError(f"Expected `task` to be callable, but got {type(task)}.")

        if not isinstance(args, Iterable) and args is not None:
            raise TypeError(f"Expected `args` to be Iterable, but got {type(args)}.")

        if not isinstance(kwargs, Mapping) and kwargs is not None:
            raise TypeError(f"Expected `kwargs` to be Mapping, but got {type(kwargs)}.")

        iid = self.__atomic.value
        unit = SyncTaskUnit(iid, name, task, args, kwargs)
        with self._lock:
            self.__task_list.append(unit)

        self.callback(TASK_SEQUENCE_ADD, unit)
        return iid

    def clear(self) -> None:
        with self._lock:
            self.__task_list.clear()

        self.callback(TASK_SEQUENCE_CLEAR)

    def __run(self) -> NoReturn:
        while True:
            try:
                self._event.wait()

                with self._lock:
                    if not self.__task_list:
                        self._event.clear()
                        self.callback(TASK_SEQUENCE_IDLE)
                        continue

                    unit = self.__task_list.pop(0)

                self.callback(TASK_SEQUENCE_EXEC, unit)
                unit.task(*unit.args, **unit.kwargs)

            except Exception as e:
                self._event.clear()
                self.clear()
                self.callback(TASK_SEQUENCE_EXCEPT, e)
                continue

    def start(self) -> bool:
        with self._lock:
            if self.thread is not None and self.thread.is_alive():
                return False

            self.thread = threading.Thread(None, self.__run, "Synchronization", (), {}, daemon=True)
            self.thread.start()

        self.callback(TASK_SEQUENCE_START)
        return True



# Licensed under the GNU Lesser General Public License v3.0.
# ezudesign Copyright (C) 2023 numlinka.

__all__ = [
    "TASK_FLOW_START",
    "TASK_FLOW_EXEC",
    "TASK_FLOW_STOP",
    "TASK_FLOW_END",
    "TASK_FLOW_EXCEPT",
    "StopTaskFlow",
    "TaskItem",
    "TaskFlowDefaultAction",
    "TaskFlow"
]

# std
import copy
import threading

from typing import Optional, Iterable, Mapping, Callable, Any
from dataclasses import dataclass

# site
from typex import Static, MultitonAtomic


TASK_FLOW_ADD = "task_flow_add"
TASK_FLOW_REMOVE = "task_flow_remove"
TASK_FLOW_START = "task_flow_start"
TASK_FLOW_EXEC = "task_flow_exec"
TASK_FLOW_STOP = "task_flow_stop"
TASK_FLOW_END = "task_flow_end"
TASK_FLOW_EXCEPT = "task_flow_except"


@dataclass(frozen=True)
class TaskItem (object):
    iid: int
    task: Callable
    priority: int
    name: str
    async_: bool


class StopTaskFlow (Exception): ...


class TaskFlowDefaultAction (Static):
    @staticmethod
    def callback(self: "TaskFlow", action: str, detail: Optional[TaskItem | Exception] = None) -> None:
        ...

    @staticmethod
    def exec_async_task(self: "TaskFlow", taskunit: TaskItem, args: Iterable[Any], kwargs: Mapping[str, Any]) -> None:
        threading.Thread(None, taskunit.task, f"{self.name} - {taskunit.name} # {taskunit.iid}", args, kwargs).start()

    @staticmethod
    def exec_sync_task(self: "TaskFlow", taskunit: TaskItem, args: Iterable[Any], kwargs: Mapping[str, Any]) -> None:
        taskunit.task(*args, **kwargs)


class TaskFlow (object):
    def __init__(self, name: Optional[str] = None) -> None:
        self._lock = threading.RLock()
        self.__name = type(self).__name__ if name is None else name
        self.__atomic = MultitonAtomic(instance_name="numlinka - ezudesign - TaskFlow")
        self.__tasks: list[TaskItem] = []

    @property
    def name(self) -> str:
        with self._lock:
            return self.__name

    def callback(self, action: str, detail: Optional[TaskItem | Exception] = None) -> None:
        TaskFlowDefaultAction.callback(self, action, detail)

    def set_name(self, name: str) -> None:
        if not isinstance(name, str):
            raise TypeError(f"Expected `name` to be str, but got {type(name)}.")

        with self._lock:
            self.__name = name

    def __sort_key(self, item: TaskItem) -> int:
        return item.priority * 2 + (0 if item.async_ else 1)

    def add_task(self, task: Callable, priority: int = 1000, name: Optional[str] = None, async_: bool = False) -> int:
        if not callable(task):
            raise TypeError(f"Expected `task` to be callable, but got {type(task)}.")

        if not isinstance(name, str) and name is not None:
            raise TypeError(f"Expected `name` to be str or None, but got {type(name)}.")

        if not isinstance(priority, int):
            raise TypeError(f"Expected `priority` to be int, but got {type(priority)}.")

        if not 0 <= priority <= 10000:
            raise ValueError(f"Expected `priority` to be in range [0, 1000], but got {priority}.")

        if not isinstance(async_, bool):
            raise TypeError(f"Expected `async_`to be bool, but got {type(async_)}.")

        if name is None:
            name = task.__name__

        iid = self.__atomic.value
        item = TaskItem(iid, task, priority, name, async_)
        self.callback(TASK_FLOW_ADD, item)

        with self._lock:
            self.__tasks.append(item)
            self.__tasks.sort(key=self.__sort_key)

        return iid

    def remove_task(self, anymark: int | str | Callable) -> bool:
        if not isinstance(anymark, (int, str)) and not callable(anymark):
            raise TypeError(f"Expected `anymark` to be int, str or callable, but got {type(anymark)}.")

        with self._lock:
            index = -1

            for i, item in enumerate(self.__tasks):
                if isinstance(anymark, int):
                    if item.iid == anymark:
                        index = i
                        break

                elif isinstance(anymark, str):
                    if item.name == anymark:
                        index = i
                        break

                else:
                    if item.task == anymark:
                        index = i
                        break

            if index == -1:
                return False

            del self.__tasks[index]
            self.callback(TASK_FLOW_REMOVE, item)
            return True

    def _exec_async_task(self, taskunit: TaskItem, args: Iterable[Any], kwargs: Mapping[str, Any]) -> None:
        self.callback(TASK_FLOW_EXEC, taskunit)
        TaskFlowDefaultAction.exec_async_task(self, taskunit, args, kwargs)

    def _exec_sync_task(self, taskunit: TaskItem, args: Iterable[Any], kwargs: Mapping[str, Any]) -> None:
        self.callback(TASK_FLOW_EXEC, taskunit)
        TaskFlowDefaultAction.exec_sync_task(self, taskunit, args, kwargs)

    def run(self, *args, **kwargs) -> None:
        self.callback(TASK_FLOW_START)
        with self._lock:
            copy_tasks = copy.copy(self.__tasks)

        try:
            for item in copy_tasks:
                if item.async_:
                    self._exec_async_task(item, args, kwargs)
                else:
                    self._exec_sync_task(item, args, kwargs)

            else:
                self.callback(TASK_FLOW_END)

        except StopTaskFlow:
            self.callback(TASK_FLOW_STOP, (item))

        except Exception as e:
            self.callback(TASK_FLOW_EXCEPT, e)
            raise e

    def __call__(self, *args, **kwargs):
        self.run(*args, **kwargs)
